/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// world.h

typedef struct
{
	vec3_t	normal;
	float	dist;
} plane_t;


// replaced by lord havocs 
// version "nice tracing" :)
typedef struct
{
	// if true, the entire trace was in solid
	qboolean	allsolid;
	// if true, the initial point was in solid
	qboolean	startsolid;
	// if true, the trace passed through empty somewhere
	qboolean	inopen;
	// if true, the trace passed through water somewhere
	qboolean	inwater;
	// fraction of the total distance that was traveled before impact
	// (1.0 = did not hit anything)
	float		fraction;
	// final position
	vec3_t		endpos;
	// surface normal at impact
	plane_t		plane;
	// entity the surface is on
	edict_t		*ent;			// entity the surface is on
	// if not zero, treats this value as empty, and all others as solid (impact
	// on content change)
	int			startcontents;
	// the contents that was hit at the end or impact point
	int			endcontents;
} trace_t;


// entities never clip against themselves, or their owner
// line of sight checks trace->crosscontent, but bullets don't
#define	MOVE_NORMAL		0
#define	MOVE_NOMONSTERS	1
#define	MOVE_MISSILE	2

typedef struct
{
	vec3_t		boxmins, boxmaxs;	// enclose the test object along entire move
	float		*mins, *maxs;		// size of the moving object
	vec3_t		mins2, maxs2;		// size when clipping against mosnters
	float		*start, *end;
	trace_t		trace;
	int			type;
	edict_t		*passedict;
} moveclip_t;

// 1/32 epsilon to keep floating point happy
#define DIST_EPSILON (0.03125)

#define HULLCHECKSTATE_EMPTY 0
#define HULLCHECKSTATE_SOLID 1
#define HULLCHECKSTATE_DONE  2

// modified a bit, now uses vec3_t instead of float[3]
typedef struct
{
	// the hull we're tracing through
	const hull_t	*hull;

	// the trace structure to fill in
	trace_t			*trace;

	// start and end of the trace (in model space)
	vec3_t			start;
	vec3_t			end;

	// end - start
	vec3_t			dist;
} RecursiveHullCheckTraceInfo_t;

void SV_ClearWorld (void);
// called after the world model has been loaded, before linking any entities

void SV_UnlinkEdict (edict_t *ent);
// call before removing an entity, and before trying to move one,
// so it doesn't clip against itself
// flags ent->v.modified

void SV_LinkEdict (edict_t *ent, qboolean touch_triggers);
// Needs to be called any time an entity changes origin, mins, maxs, or solid
// flags ent->v.modified
// sets ent->v.absmin and ent->v.absmax
// if touchtriggers, calls prog functions for the intersected triggers

int SV_HullPointContents (hull_t *hull, int num, vec3_t p);
int SV_TruePointContents (vec3_t p);
int SV_PointContents (vec3_t p);
// returns the CONTENTS_* value from the world at the given point.
// does not check any entities at all
// the non-true version remaps the water current contents to content_water

qboolean SV_RecursiveHullCheck (hull_t *hull, int num, float p1f, float p2f, vec3_t p1, vec3_t p2, trace_t *trace);
edict_t	*SV_TestEntityPosition (edict_t *ent);

trace_t SV_Move (vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end, int type, edict_t *passedict);
// mins and maxs are reletive

// if the entire move stays in a solid volume, trace.allsolid will be set

// if the starting point is in a solid, it will be allowed to move out
// to an open area

// nomonsters is used for line of sight or edge testing, where mosnters
// shouldn't be considered solid objects

// passedict is explicitly excluded from clipping checks (normally NULL)
